/* Copyright 2015 Research Foundation State University of New York   */

/* This file is part of QUB Express.                                      */

/* QUB Express is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Express is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

#ifndef QUB_MEAN_H
#define QUB_MEAN_H

#include "qubfast.h"

#ifdef __cplusplus
extern "C" {
#endif

QUBFAST_API void    qub_mean_rss(float *data, int N, double *mean, double *rss);

QUBFAST_API void*   qub_mean_rss_init();
QUBFAST_API void    qub_mean_rss_free(void *obj);
QUBFAST_API void    qub_mean_rss_add(void *obj, float *data, int N);
QUBFAST_API int     qub_mean_rss_read(void *obj, double *mean, double *rss);

#ifdef __cplusplus
}


class IncrementalRSS {
  // Object which holds the mean, residual sum of squares (rss), and count (n)
  // of a collection of numbers,
  // and can add and remove a number from the collection in constant time.
public:
  double mean;
  double rss;
  double n;
  IncrementalRSS()
    : mean(0.0), rss(0.0), n(0.0)
  {}
  IncrementalRSS(double _mean, double _rss, double _n)
    : mean(_mean), rss(_rss), n(_n)
  {}
  void set(double _mean, double _rss, double _n) {
    mean = _mean;
    rss = _rss;
    n = _n;
  }
  template<class T>
  void add(T x) {
    double last_mean = mean;
    n += 1.0;
    mean = (1.0/n)*((n-1)*last_mean + x);
    double dm = mean - last_mean;
    double d = x - last_mean;
    rss += d*d - n*dm*dm;
  }
  template<class T>
  void add_n(T* data, long count) {
    double last_mean, this_mean=mean;
    double this_rss=rss;
    double this_n = n;
    while ( count-- ) {
      double x = *(data++);
      if ( this_n == 0 ) {
	this_mean = x;
	this_n = 1;
      } else {
	last_mean = this_mean;
	this_n += 1.0;
	this_mean = (1.0/this_n)*((this_n-1)*last_mean + x);
	double dm = this_mean - last_mean;
	double d = x - last_mean;
	this_rss += d*d - this_n*dm*dm;
      }
    }
    mean = this_mean;
    rss = this_rss;
    n = this_n;
  }
  template<class T>
  void remove(T x) {
    double last_mean = mean;
    mean = (1.0/(n-1.0))*(n*last_mean - x);
    double d = x - mean;
    double dm = mean - last_mean;
    rss += - d*d + n*dm*dm;
    n -= 1.0;
  }
  double std() const {
    return n ? sqrt(rss/n) : 0.0;
  }
  double sample_std() const {
    return (n>=2.0) ? sqrt(rss/(n-1)) : 0.0;
  }
  const IncrementalRSS& operator+=(const IncrementalRSS& other) {
    double last_mean = mean;
    double last_n = n;
    n += other.n;
    mean = last_n*last_mean/n + other.n*other.mean/n;
    rss += other.rss + last_n*squared(last_mean - mean) + other.n*squared(other.mean - mean);
    return *this;
  }
  const IncrementalRSS operator+(const IncrementalRSS& other) const {
    // returns the composite mean and rss; see "How to calculate composite standard deviations.html"
    IncrementalRSS result(mean, rss, n);
    result += other;
    return result;
  }
};

#endif

#endif
